package org.spagic.utils;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.stp.im.BasicProperty;
import org.eclipse.stp.im.ConfigurableElement;
import org.eclipse.stp.im.Process;
import org.eclipse.stp.im.Property;
import org.eclipse.stp.im.ServiceBinding;
import org.eclipse.stp.im.Step;
import org.eclipse.stp.im.StpIntermediateModel;
import org.eclipse.stp.im.Transition;
import org.eclipse.stp.im.constants.ImConstants;


public class SpagicJBIImPreprocessor extends AbstractImPreprocessor {

	private static final String IS_STEP_INTERMEDIATE_EVENT = "im.isStepIntermediateEvent";
	private static final String JBI_TECHNOLOGY = "JBI";
	
	private String runtimeId = null;

	
	public SpagicJBIImPreprocessor(String runtimeId){
		super();
		this.runtimeId = runtimeId;
	}
	
	public void process(StpIntermediateModel im) {
		for (Process process : im.getProcessCollection().getProcesses()) {
			if (!SpagicUtils.isProcessDeployableForRuntime(process, getRuntimeId()))
				continue;
			handleImProcess(im, process);
		}
	}
	
	public void handleImProcess(StpIntermediateModel im, Process process) {
	
		// Create the technology property
		createTechnologyProperty(process, JBI_TECHNOLOGY);
		
		// Handle The Synchronizer Pattern
		handleSynchronizer(im, process);
		
		// Handle Pipeline
		handlePipeline(im, process);
		
		// Handle Wait Event
		handleWaitEvent(im, process);
			
	}
		
	public void createServiceBindingForRuntime(StpIntermediateModel im, String serviceBindingName) {
		
		ServiceBinding serviceBinding = factory.createServiceBinding();
		serviceBinding.setName(serviceBindingName);
		BasicProperty runtimeIdProperty = factory.createBasicProperty();
		runtimeIdProperty.setKey(ImConstants.IM_STEP_BINDING_RUNTIME_ID);
		runtimeIdProperty.setValue(getRuntimeId());
		serviceBinding.getProperties().put(runtimeIdProperty.getKey(),
				runtimeIdProperty);
		im.getServicebindings().add(serviceBinding);
		
	}

	public ServiceBinding getRightStepBinding(Step s) {
	Property runtimPropertyForServiceBinding = null;
	for (ServiceBinding sb : s.getStepbindings()){
		runtimPropertyForServiceBinding = sb.getProperties().get(ImConstants.IM_STEP_BINDING_RUNTIME_ID);
		if (runtimPropertyForServiceBinding != null){
			if (runtimPropertyForServiceBinding.getValue().equals(getRuntimeId())){
				return sb;
			}
		}else{
			SpagicLogger.warning("ServiceBinding["+sb.getName()+"] -> for Step["+s.getName()+"] does not have a property called "+ ImConstants.IM_POOL_RUNTIME_ID);
		}
	}
	SpagicLogger.warning("Not found for Step["+s.getName()+"] a binding for runtime "+ ImConstants.IM_POOL_RUNTIME_ID);
	return null;
	}

	public String getRuntimeId() {
		return runtimeId;
	}

	public void setRuntimeId(String runtimeId) {
		this.runtimeId = runtimeId;
	}
	
	public String getBasicPropertyValue( ConfigurableElement ce, String propertyName ){
		Property p = ce.getProperties().get(propertyName);
		if (p != null){
			 if (p instanceof BasicProperty) {
				return ((BasicProperty)p).getValue();	
			}else{
				return p.toString();
			}
		}else{
			return null;
		}
	}
	public void handleWaitEvent(StpIntermediateModel im, Process aProcess ){
		List<Step> stepToAddToProcess = new ArrayList<Step>();
		List<Step> replyStepsToRemove = new ArrayList<Step>();
		
		for (Step aStep : aProcess.getSteps()) {
			if ( isServiceBinding(aStep, "JBI-HumanTaskEvent") || isServiceBinding(aStep,"JBI-WaitExternalEvent")) {
				
				if (aStep.getSourceTransitions().size() > 1){
					throw new RuntimeException(" A Wait or HumanTaskEvent Binding could have only one destination ");
				}
				
				Transition destinationAfterWaitTransition = getFirstOutcomingTransition(aStep);
				if (destinationAfterWaitTransition == null)
					throw new RuntimeException(" A Wait or HumanTaskEvent Service Must have one destination ");
				
				Step destinationAfterWaitStep = destinationAfterWaitTransition.getTarget();
				
				String signalStepName = aStep.getName() + "Signal";
				
				// Create Signal Step
				Step signalStep = createStep(im, signalStepName, "JBI-SignalEvent");
				stepToAddToProcess.add(signalStep);
				
				String destinationURI = SpagicUtils.getStepNameInProcess(aProcess, destinationAfterWaitStep);
				String signalURI = SpagicUtils.getStepNameInProcess(aProcess, signalStep);
				
				setPropertyOnStep(aStep, "destinationURI", destinationURI);
				setPropertyOnStep(aStep, "signalEndpointURI", signalURI);
				
				setPropertyOnStep(signalStep, "destinationURI", destinationURI);
				
				
				destinationAfterWaitTransition.setSource(signalStep);
				
				createTransitionAndAddToProcess(aProcess, aStep, signalStep);
				
				Transition incomingFromExternalBindingTransition = findTheTransitionComingFromExternalEventHandler(aStep);				
				if ( incomingFromExternalBindingTransition == null ){
					 
					 if (isServiceBinding(aStep,"JBI-WaitExternalEvent")) {
						 SpagicLogger.error(" When using a Wait External Event In The Process must be present a service binding belonging to External Event Service");
					 }else if (isServiceBinding(aStep,"JBI-HumanTaskEvent")) {
						 setPropertyOnStep(aStep, "EventType", "GenericEvent");
					 }
					 
				} else {
					
					if (isServiceBinding(aStep,"JBI-HumanTaskEvent")) {
						 SpagicLogger.error(" When using a JBI-HumanTaskEvent no External Event Service is Required");
					}else if (isServiceBinding(aStep,"JBI-WaitExternalEvent")) {
						 
						 Step incomingExternalServiceStep = incomingFromExternalBindingTransition.getSource();
						 String handledEventType = getEventTypeHandled(incomingExternalServiceStep);
						 setPropertyOnStep(aStep, "EventType",handledEventType );
						 setSignalEndpointURIOnStep(aProcess, incomingExternalServiceStep, signalStep);
						 
						 // Create The Matching Engine Step
						 String eventMatcherServiceBindingName = "JBI-EventMatcher";
						 Step eventMatcherStep = createStep(im, aStep.getName() + "EventMatcher", eventMatcherServiceBindingName);
						 setPropertyOnStep(eventMatcherStep, "signalEndpointURI", signalURI);
						 stepToAddToProcess.add(eventMatcherStep);
						 
						 
						 // Copy on Event Matcher the Matching rules
						 String numberOfRulesOnEventAttributeString = getBasicPropertyValue(incomingFromExternalBindingTransition, "NumberOfRulesOnEventAttribute");
						 String numberOfRulesOnEventPayloadString = getBasicPropertyValue(incomingFromExternalBindingTransition, "NumberOfRulesOnEventPayload");
						 
						 
						 
						 if (numberOfRulesOnEventAttributeString != null){
							 int numberOfRulesOnEventAttribute = Integer.valueOf(numberOfRulesOnEventAttributeString);
							 setPropertyOnStep(eventMatcherStep, "NumberOfRulesOnEventAttribute", numberOfRulesOnEventAttributeString);
							 
							 String pName,  pValue = null;
							 for ( int i = 0; i < numberOfRulesOnEventAttribute; i++){
								 pName = MessageFormat.format("RulesOnEventAttibute[{0}].WaitInstanceAttributeName",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
								 pName = MessageFormat.format("RulesOnEventAttibute[{0}].EventAttributeName",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
									
							 }
						 }
						 
						 if (numberOfRulesOnEventPayloadString != null){
							 int numberOfRulesOnEventPayload = Integer.valueOf(numberOfRulesOnEventPayloadString);
							 setPropertyOnStep(eventMatcherStep, "NumberOfRulesOnEventPayload", numberOfRulesOnEventPayloadString);
							 
							 String pName,  pValue = null;
							 for ( int i = 0; i < numberOfRulesOnEventPayload; i++){
								 pName = MessageFormat.format("RulesOnEventPayload[{0}].WaitInstanceAttributeName",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
								 pName = MessageFormat.format("RulesOnEventPayload[{0}].XPathExpression",i);
								 pValue =  getBasicPropertyValue(incomingFromExternalBindingTransition,pName);
								 setPropertyOnStep(eventMatcherStep, pName, pValue);
							 }
						 }
					     
					     // Relocating Transitions
						 if (handledEventType.equalsIgnoreCase("WebserviceEvent")){

							 String eventWrapperStepName = "JBI-WrapToEvent";
							 Step eventWrapperStep = createStep(im, incomingExternalServiceStep.getName() + "EventWrapper", eventWrapperStepName);
							 setPropertyOnStep(eventWrapperStep, "EventType",handledEventType );
							 String eventMatcherURI = SpagicUtils.getStepNameInProcess(aProcess, eventMatcherStep);
							 setPropertyOnStep(eventWrapperStep, "eventMatcherURI", eventMatcherURI);
							 setPropertyOnStep(eventWrapperStep, IS_STEP_INTERMEDIATE_EVENT, "true");
							 stepToAddToProcess.add(eventWrapperStep);
							 incomingFromExternalBindingTransition.setTarget(eventWrapperStep);
							 createTransitionAndAddToProcess(aProcess, eventWrapperStep, eventMatcherStep);
							 createTransitionAndAddToProcess(aProcess, eventMatcherStep, signalStep);
							 
						 }else{
							 incomingFromExternalBindingTransition.setTarget(eventMatcherStep);
							 createTransitionAndAddToProcess(aProcess, eventMatcherStep, signalStep);
						 }
					}
					
				} //if ( incomingFromExternalBindingComponent == null )
			} // if ( isServiceBinding(aStep, "JBI-HumanTaskEvent") || isServiceBinding(aStep,"JBI-WaitExternalEvent"))
		}
		aProcess.getSteps().removeAll(replyStepsToRemove);		
		aProcess.getSteps().addAll(stepToAddToProcess);
	} //public void handleWaitEvent(StpIntermediateModel im, Process aProcess )

	public void handlePipeline(StpIntermediateModel im, Process aProcess ){

		List<Step> stepToAddToProcess = new ArrayList<Step>();
		List<Step> replyStepsToRemove = new ArrayList<Step>();
		
		Step transformerStep = null;
		for (Step pipelineStep : aProcess.getSteps()) {
			if (isServiceBinding(pipelineStep, "JBI-WebServicePipeline") || isServiceBinding(pipelineStep,"JBI-TCPIP-Pipeline") || isServiceBinding(pipelineStep,"JBI-CXF-Pipeline")) {
			
				String serviceBindingName = null;
				if (isServiceBinding(pipelineStep, "JBI-WebServicePipeline")){
					serviceBindingName = "JBI-HTTPInputOutputBindingComponent";
				}
				
				if (isServiceBinding(pipelineStep, "JBI-TCPIP-Pipeline")){
					serviceBindingName = "JBI-TcpIp-InputOutputBindingComponent";
				}
				
				if (isServiceBinding(pipelineStep, "JBI-CXF-Pipeline")){
					serviceBindingName="JBI-CXF-InputOutputBindingComponent";
				}
				
				transformerStep = createStep(im, pipelineStep.getName()
						+ SpagicUtilsConstants.JBI_TRANSFORMER_SUFFIX, serviceBindingName);
				stepToAddToProcess.add(transformerStep);
				
				copyAllProperties(pipelineStep, transformerStep);
				

				setPropertyOnStep(pipelineStep, "transformerStepName", transformerStep.getName());
				setPropertyOnStep(transformerStep, "IsTransformerPipeline", "true");
				
				createTransitionAndAddToProcess(aProcess, pipelineStep, transformerStep);				
				createTransitionAndAddToProcess(aProcess, transformerStep, pipelineStep);

			}
		}// end
		aProcess.getSteps().removeAll(replyStepsToRemove);		
		aProcess.getSteps().addAll(stepToAddToProcess);
	}
	
	
	public boolean isCandidateForSynchronizer(Step step){
		
		ServiceBinding sb  = getRightStepBinding(step);
		return (sb != null) && 
			( isServiceBinding(sb,"JBI-HTTPInputOutputBindingComponent") || isServiceBinding(sb, "JBI-TcpIp-InputOutputBindingComponent") || isServiceBinding(sb, "JBI-CXF-InputOutputBindingComponent"));
	}
	
	public Step createSynchronizerStepForStartStep(StpIntermediateModel im, Step sourceOfSynchronizer){
		return createStep(im, sourceOfSynchronizer.getName()
				+ SpagicUtilsConstants.JBI_SYNC_SUFFIX, "JBI-Sync");
	}
	
	///
	public  void handleSynchronizer(StpIntermediateModel im,  Process process) {
		boolean syncAdded = false;
		
		List<Step> stepToAddToProcess = new ArrayList<Step>();
		List<Step> replyStepsToRemove = new ArrayList<Step>();
		
	
		Transition sourceTransition = null;
		Step destinationStep = null;
		Step synchronizerStep = null;
		createServiceBindingForRuntime(im, "JBI-Sync");
		for (Step aStep : process.getSteps()) {
			
			if (isCandidateForSynchronizer(aStep)){
					
				if (stepNumber(process) > 3){
				
					// Get The first outcoming transition to save the destination Step
					sourceTransition = getFirstOutcomingTransition(aStep);
					destinationStep = sourceTransition.getTarget();
				
					// Removing Original Transition
					removeTransitionFromProcess(process,sourceTransition);
					removeFromStepOutcomingTransition(aStep,sourceTransition);
					removeFromStepIncomingTransition(destinationStep,sourceTransition);
				
					// Creating The Synchnronizer Step
					synchronizerStep = createSynchronizerStepForStartStep(im, aStep);
					stepToAddToProcess.add(synchronizerStep);

					// Relocate Transition
					createTransitionAndAddToProcess(process, aStep, synchronizerStep);
					createTransitionAndAddToProcess(process, synchronizerStep, destinationStep);
					createTransitionAndAddToProcess(process, synchronizerStep, aStep);
				
					syncAdded = true;
				
				}else{
					// No need to create a Sync relocate only the transition
					sourceTransition = getFirstOutcomingTransition(aStep);
					destinationStep = sourceTransition.getTarget();
					createTransitionAndAddToProcess(process, destinationStep, aStep);
					syncAdded = false;
				}
			}
		}// for Step Loop 
	
	
		// Change the transitions of the steps that has target reply to sync
		for (Step step : process.getSteps()) {
			List<Transition> transitionsToRemoveFromOutcomingTransition = new ArrayList<Transition>();
			for (Transition tr : step.getSourceTransitions()) {
				if (isServiceBinding(tr.getTarget(), "JBI-Reply")){
					replyStepsToRemove.add(tr.getTarget());
					if (syncAdded)
						tr.setTarget(synchronizerStep);
					else{
						transitionsToRemoveFromOutcomingTransition.add(tr);
					}
				}
			}//for (Transition tr :
			removeFromStepOutcomingTransition(step,transitionsToRemoveFromOutcomingTransition);
			process.getTransitions().removeAll(transitionsToRemoveFromOutcomingTransition);
		}//for (Step aStep :
		
		process.getSteps().removeAll(replyStepsToRemove);		
		process.getSteps().addAll(stepToAddToProcess);
	}
	
	////////

	public String getEventTypeHandled(Step aStep){
		
		if (isServiceBinding(aStep, "FileEventBindingComponent") || isServiceBinding(aStep, "FTPEventBindingComponent")){
			return "FileEvent";
		}else if (isServiceBinding(aStep, "WebserviceEventBindingComponent")){
			return "WebserviceEvent";
		}
		return "GenericEvent";
		
	}
	public Transition findTheTransitionComingFromExternalEventHandler(Step aStep){
		
		List<Transition> incomingTransitions = aStep.getTargetTransitions();
		for (Transition incomingTransition : incomingTransitions){
			if (isIncomingTransitionFromExternalEventHandlerBindingComponent(incomingTransition))
				return incomingTransition;
		}
		return null;
	}
	
	public void setSignalEndpointURIOnStep(Process process, Step targetStep, Step signalStep){
		setPropertyOnStep(targetStep, "signalEndpointURI", SpagicUtils.getStepNameInProcess(process, signalStep));
	}
	
	public boolean isServiceBinding(Step step, String serviceBindingName) {
		return isServiceBinding(getRightStepBinding(step), serviceBindingName);
	}
	
}
